@@ -53,6 +53,7 @@
 #include "shared/image-loader.h"
 #include "shared/os-compatibility.h"
 #include "shared/cairo-util.h"
+#include "shared/timespec-util.h"
 #include "fullscreen-shell-unstable-v1-client-protocol.h"
 #include "xdg-shell-unstable-v6-client-protocol.h"
 #include "presentation-time-server-protocol.h"
@@ -111,7 +112,7 @@ struct wayland_output {
 
 	int keyboard_count;
 
-	char *name;
+	char *title;
 	struct frame *frame;
 
 	struct {
@@ -130,13 +131,13 @@ struct wayland_output {
 	} shm;
 
 	struct weston_mode mode;
-	uint32_t scale;
 
 	struct wl_callback *frame_cb;
 };
 
 struct wayland_parent_output {
-	struct wayland_output *output;
+	struct wayland_backend *backend;	/**< convenience */
+	struct wayland_head *head;
 	struct wl_list link;
 
 	struct wl_output *global;
@@ -153,11 +154,18 @@ struct wayland_parent_output {
 	uint32_t transform;
 	uint32_t scale;
 
+	struct wl_callback *sync_cb;	/**< wl_output < 2 done replacement */
+
 	struct wl_list mode_list;
 	struct weston_mode *preferred_mode;
 	struct weston_mode *current_mode;
 };
 
+struct wayland_head {
+	struct weston_head base;
+	struct wayland_parent_output *parent_output;
+};
+
 struct wayland_shm_buffer {
 	struct wayland_output *output;
 	struct wl_list link;
@@ -166,7 +174,7 @@ struct wayland_shm_buffer {
 	struct wl_buffer *buffer;
 	void *data;
 	size_t size;
-	pixman_region32_t damage;
+	pixman_region32_t damage;		/**< in global coords */
 	int frame_damaged;
 
 	pixman_image_t *pm_image;
@@ -190,6 +198,8 @@ struct wayland_input {
 		} cursor;
 	} parent;
 
+	struct weston_touch_device *touch_device;
+
 	enum weston_key_state_update keyboard_state_update;
 	uint32_t key_serial;
 	uint32_t enter_serial;
@@ -207,6 +217,12 @@ struct wayland_input {
 
 struct gl_renderer_interface *gl_renderer;
 
+static inline struct wayland_head *
+to_wayland_head(struct weston_head *base)
+{
+	return container_of(base, struct wayland_head, base);
+}
+
 static inline struct wayland_output *
 to_wayland_output(struct weston_output *base)
 {
@@ -301,7 +317,7 @@ wayland_output_get_shm_buffer(struct wayland_output *output)
 	if (sb == NULL) {
 		weston_log("could not zalloc %zu memory for sb: %m\n", sizeof *sb);
 		close(fd);
-		free(data);
+		munmap(data, height * stride);
 		return NULL;
 	}
 
@@ -309,8 +325,8 @@ wayland_output_get_shm_buffer(struct wayland_output *output)
 	wl_list_init(&sb->free_link);
 	wl_list_insert(&output->shm.buffers, &sb->link);
 
-	pixman_region32_init_rect(&sb->damage, 0, 0,
-				  output->base.width, output->base.height);
+	pixman_region32_init(&sb->damage);
+	pixman_region32_copy(&sb->damage, &output->base.region);
 	sb->frame_damaged = 1;
 
 	sb->data = data;
@@ -488,7 +504,8 @@ wayland_output_start_repaint_loop(struct weston_output *output_base)
 #ifdef ENABLE_EGL
 static int
 wayland_output_repaint_gl(struct weston_output *output_base,
-			  pixman_region32_t *damage)
+			  pixman_region32_t *damage,
+			  void *repaint_data)
 {
 	struct wayland_output *output = to_wayland_output(output_base);
 	struct weston_compositor *ec = output->base.compositor;
@@ -555,11 +572,15 @@ wayland_shm_buffer_attach(struct wayland_shm_buffer *sb)
 	int i, n;
 
 	pixman_region32_init(&damage);
+	pixman_region32_copy(&damage, &sb->damage);
+	pixman_region32_translate(&damage, -sb->output->base.x,
+				  -sb->output->base.y);
+
 	weston_transformed_region(sb->output->base.width,
 				  sb->output->base.height,
 				  sb->output->base.transform,
 				  sb->output->base.current_scale,
-				  &sb->damage, &damage);
+				  &damage, &damage);
 
 	if (sb->output->frame) {
 		frame_interior(sb->output->frame, &ix, &iy, &iwidth, &iheight);
@@ -595,7 +616,8 @@ wayland_shm_buffer_attach(struct wayland_shm_buffer *sb)
 
 static int
 wayland_output_repaint_pixman(struct weston_output *output_base,
-			      pixman_region32_t *damage)
+			      pixman_region32_t *damage,
+			      void *repaint_data)
 {
 	struct wayland_output *output = to_wayland_output(output_base);
 	struct wayland_backend *b =
@@ -636,16 +658,25 @@ wayland_output_repaint_pixman(struct weston_output *output_base,
 static void
 wayland_backend_destroy_output_surface(struct wayland_output *output)
 {
-	if (output->parent.xdg_toplevel)
+	assert(output->parent.surface);
+
+	if (output->parent.xdg_toplevel) {
 		zxdg_toplevel_v6_destroy(output->parent.xdg_toplevel);
+		output->parent.xdg_toplevel = NULL;
+	}
 
-	if (output->parent.xdg_surface)
+	if (output->parent.xdg_surface) {
 		zxdg_surface_v6_destroy(output->parent.xdg_surface);
+		output->parent.xdg_surface = NULL;
+	}
 
-	if (output->parent.shell_surface)
+	if (output->parent.shell_surface) {
 		wl_shell_surface_destroy(output->parent.shell_surface);
+		output->parent.shell_surface = NULL;
+	}
 
 	wl_surface_destroy(output->parent.surface);
+	output->parent.surface = NULL;
 }
 
 static void
@@ -701,11 +732,12 @@ wayland_output_destroy(struct weston_output *base)
 
 	wayland_output_disable(&output->base);
 
-	weston_output_destroy(&output->base);
+	weston_output_release(&output->base);
 
 	if (output->frame_cb)
 		wl_callback_destroy(output->frame_cb);
 
+	free(output->title);
 	free(output);
 }
 
@@ -752,7 +784,8 @@ cleanup_window:
 static int
 wayland_output_init_pixman_renderer(struct wayland_output *output)
 {
-	return pixman_renderer_output_create(&output->base);
+	return pixman_renderer_output_create(&output->base,
+				     PIXMAN_RENDERER_OUTPUT_USE_SHADOW);
 }
 
 static void
@@ -776,6 +809,14 @@ wayland_output_resize_surface(struct wayland_output *output)
 		wl_surface_set_input_region(output->parent.surface, region);
 		wl_region_destroy(region);
 
+		if (output->parent.xdg_surface) {
+			zxdg_surface_v6_set_window_geometry(output->parent.xdg_surface,
+							    ix,
+							    iy,
+							    iwidth,
+							    iheight);
+		}
+
 		frame_opaque_rect(output->frame, &ix, &iy, &iwidth, &iheight);
 		region = wl_compositor_create_region(b->parent.compositor);
 		wl_region_add(region, ix, iy, iwidth, iheight);
@@ -794,6 +835,14 @@ wayland_output_resize_surface(struct wayland_output *output)
 		wl_region_add(region, 0, 0, width, height);
 		wl_surface_set_opaque_region(output->parent.surface, region);
 		wl_region_destroy(region);
+
+		if (output->parent.xdg_surface) {
+			zxdg_surface_v6_set_window_geometry(output->parent.xdg_surface,
+							    0,
+							    0,
+							    width,
+							    height);
+		}
 	}
 
 #ifdef ENABLE_EGL
@@ -833,36 +882,17 @@ wayland_output_set_windowed(struct wayland_output *output)
 {
 	struct wayland_backend *b =
 		to_wayland_backend(output->base.compositor);
-	int tlen;
-	char *title;
 
 	if (output->frame)
 		return 0;
 
-	if (output->name) {
-		tlen = strlen(output->name) + strlen(WINDOW_TITLE " - ");
-		title = malloc(tlen + 1);
-		if (!title)
-			return -1;
-
-		snprintf(title, tlen + 1, WINDOW_TITLE " - %s", output->name);
-	} else {
-		title = strdup(WINDOW_TITLE);
-	}
-
-	if (output->parent.xdg_toplevel)
-		zxdg_toplevel_v6_set_title(output->parent.xdg_toplevel, title);
-
 	if (!b->theme) {
 		b->theme = theme_create();
-		if (!b->theme) {
-			free(title);
+		if (!b->theme)
 			return -1;
-		}
 	}
 	output->frame = frame_create(b->theme, 100, 100,
-				     FRAME_BUTTON_CLOSE, title);
-	free(title);
+	                             FRAME_BUTTON_CLOSE, output->title, NULL);
 	if (!output->frame)
 		return -1;
 
@@ -871,8 +901,13 @@ wayland_output_set_windowed(struct wayland_output *output)
 
 	wayland_output_resize_surface(output);
 
-	if (output->parent.shell_surface)
+	if (output->parent.xdg_toplevel) {
+		zxdg_toplevel_v6_unset_fullscreen(output->parent.xdg_toplevel);
+	} else if (output->parent.shell_surface) {
 		wl_shell_surface_set_toplevel(output->parent.shell_surface);
+	} else {
+		abort();
+	}
 
 	return 0;
 }
@@ -882,9 +917,6 @@ wayland_output_set_fullscreen(struct wayland_output *output,
 			      enum wl_shell_surface_fullscreen_method method,
 			      uint32_t framerate, struct wl_output *target)
 {
-	struct wayland_backend *b =
-		to_wayland_backend(output->base.compositor);
-
 	if (output->frame) {
 		frame_destroy(output->frame);
 		output->frame = NULL;
@@ -897,10 +929,8 @@ wayland_output_set_fullscreen(struct wayland_output *output,
 	} else if (output->parent.shell_surface) {
 		wl_shell_surface_set_fullscreen(output->parent.shell_surface,
 						method, framerate, target);
-	} else if (b->parent.fshell) {
-		zwp_fullscreen_shell_v1_present_surface(b->parent.fshell,
-							output->parent.surface,
-							method, target);
+	} else {
+		abort();
 	}
 }
 
@@ -971,6 +1001,38 @@ struct zwp_fullscreen_shell_mode_feedback_v1_listener mode_feedback_listener = {
 	mode_feedback_cancelled,
 };
 
+static enum mode_status
+wayland_output_fullscreen_shell_mode_feedback(struct wayland_output *output,
+					      struct weston_mode *mode)
+{
+	struct wayland_backend *b = to_wayland_backend(output->base.compositor);
+	struct zwp_fullscreen_shell_mode_feedback_v1 *mode_feedback;
+	enum mode_status mode_status;
+	int ret = 0;
+
+	mode_feedback =
+		zwp_fullscreen_shell_v1_present_surface_for_mode(b->parent.fshell,
+								 output->parent.surface,
+								 output->parent.output,
+								 mode->refresh);
+
+	zwp_fullscreen_shell_mode_feedback_v1_add_listener(mode_feedback,
+							   &mode_feedback_listener,
+							   &mode_status);
+
+	output->parent.draw_initial_frame = false;
+	draw_initial_frame(output);
+	wl_surface_commit(output->parent.surface);
+
+	mode_status = MODE_STATUS_UNKNOWN;
+	while (mode_status == MODE_STATUS_UNKNOWN && ret >= 0)
+		ret = wl_display_dispatch(b->parent.wl_display);
+
+	zwp_fullscreen_shell_mode_feedback_v1_destroy(mode_feedback);
+
+	return mode_status;
+}
+
 static int
 wayland_output_switch_mode(struct weston_output *output_base,
 			   struct weston_mode *mode)
@@ -979,9 +1041,7 @@ wayland_output_switch_mode(struct weston_output *output_base,
 	struct wayland_backend *b;
 	struct wl_surface *old_surface;
 	struct weston_mode *old_mode;
-	struct zwp_fullscreen_shell_mode_feedback_v1 *mode_feedback;
 	enum mode_status mode_status;
-	int ret = 0;
 
 	if (output_base == NULL) {
 		weston_log("output is NULL.\n");
@@ -1015,25 +1075,11 @@ wayland_output_switch_mode(struct weston_output *output_base,
 	/* Blow the old buffers because we changed size/surfaces */
 	wayland_output_resize_surface(output);
 
-	mode_feedback =
-		zwp_fullscreen_shell_v1_present_surface_for_mode(b->parent.fshell,
-								 output->parent.surface,
-								 output->parent.output,
-								 mode->refresh);
-	zwp_fullscreen_shell_mode_feedback_v1_add_listener(mode_feedback,
-							   &mode_feedback_listener,
-							   &mode_status);
+	mode_status = wayland_output_fullscreen_shell_mode_feedback(output, mode);
 
 	/* This should kick-start things again */
-	output->parent.draw_initial_frame = true;
 	wayland_output_start_repaint_loop(&output->base);
 
-	mode_status = MODE_STATUS_UNKNOWN;
-	while (mode_status == MODE_STATUS_UNKNOWN && ret >= 0)
-		ret = wl_display_dispatch(b->parent.wl_display);
-
-	zwp_fullscreen_shell_mode_feedback_v1_destroy(mode_feedback);
-
 	if (mode_status == MODE_STATUS_FAIL) {
 		output->base.current_mode = old_mode;
 		wl_surface_destroy(output->parent.surface);
@@ -1097,6 +1143,13 @@ handle_xdg_toplevel_configure(void *data, struct zxdg_toplevel_v6 *toplevel,
 static void
 handle_xdg_toplevel_close(void *data, struct zxdg_toplevel_v6 *xdg_toplevel)
 {
+	struct wayland_output *output = data;
+	struct weston_compositor *compositor = output->base.compositor;
+
+	wayland_output_destroy(&output->base);
+
+	if (wl_list_empty(&compositor->output_list))
+		weston_compositor_exit(compositor);
 }
 
 static const struct zxdg_toplevel_v6_listener xdg_toplevel_listener = {
@@ -1109,6 +1162,8 @@ wayland_backend_create_output_surface(struct wayland_output *output)
 {
 	struct wayland_backend *b = to_wayland_backend(output->base.compositor);
 
+	assert(!output->parent.surface);
+
 	output->parent.surface =
 		wl_compositor_create_surface(b->parent.compositor);
 	if (!output->parent.surface)
@@ -1130,6 +1185,8 @@ wayland_backend_create_output_surface(struct wayland_output *output)
 		zxdg_toplevel_v6_add_listener(output->parent.xdg_toplevel,
 					      &xdg_toplevel_listener, output);
 
+		zxdg_toplevel_v6_set_title(output->parent.xdg_toplevel, output->title);
+
 		wl_surface_commit(output->parent.surface);
 
 		output->parent.wait_for_configure = true;
@@ -1162,6 +1219,7 @@ wayland_output_enable(struct weston_output *base)
 {
 	struct wayland_output *output = to_wayland_output(base);
 	struct wayland_backend *b = to_wayland_backend(base->compositor);
+	enum mode_status mode_status;
 	int ret = 0;
 
 	weston_log("Creating %dx%d wayland output at (%d, %d)\n",
@@ -1199,28 +1257,23 @@ wayland_output_enable(struct weston_output *base)
 	output->base.switch_mode = wayland_output_switch_mode;
 
 	if (b->sprawl_across_outputs) {
-		wayland_output_set_fullscreen(output,
-					      WL_SHELL_SURFACE_FULLSCREEN_METHOD_DRIVER,
-					      output->mode.refresh, output->parent.output);
+		if (b->parent.fshell) {
+			wayland_output_resize_surface(output);
 
-		if (output->parent.xdg_toplevel) {
-			zxdg_toplevel_v6_set_fullscreen(output->parent.xdg_toplevel,
-							output->parent.output);
-		}
-		else if (output->parent.shell_surface) {
-			wl_shell_surface_set_fullscreen(output->parent.shell_surface,
-							WL_SHELL_SURFACE_FULLSCREEN_METHOD_DRIVER,
-							output->mode.refresh, output->parent.output);
-		} else if (b->parent.fshell) {
-			zwp_fullscreen_shell_v1_present_surface(b->parent.fshell,
-								output->parent.surface,
-								ZWP_FULLSCREEN_SHELL_V1_PRESENT_METHOD_CENTER,
-								output->parent.output);
-			zwp_fullscreen_shell_mode_feedback_v1_destroy(
-				zwp_fullscreen_shell_v1_present_surface_for_mode(b->parent.fshell,
-										 output->parent.surface,
-										 output->parent.output,
-										 output->mode.refresh));
+			mode_status = wayland_output_fullscreen_shell_mode_feedback(output, &output->mode);
+
+			if (mode_status == MODE_STATUS_FAIL) {
+				zwp_fullscreen_shell_v1_present_surface(b->parent.fshell,
+									output->parent.surface,
+									ZWP_FULLSCREEN_SHELL_V1_PRESENT_METHOD_CENTER,
+									output->parent.output);
+
+				output->parent.draw_initial_frame = true;
+			}
+		} else {
+			wayland_output_set_fullscreen(output,
+						      WL_SHELL_SURFACE_FULLSCREEN_METHOD_DRIVER,
+						      output->mode.refresh, output->parent.output);
 		}
 	} else if (b->fullscreen) {
 		wayland_output_set_fullscreen(output, 0, 0, NULL);
@@ -1236,10 +1289,65 @@ err_output:
 	return -1;
 }
 
-static struct wayland_output *
-wayland_output_create_common(void)
+static int
+wayland_output_setup_for_parent_output(struct wayland_output *output,
+				       struct wayland_parent_output *poutput);
+
+static int
+wayland_output_setup_fullscreen(struct wayland_output *output,
+				struct wayland_head *head);
+
+static int
+wayland_output_attach_head(struct weston_output *output_base,
+			   struct weston_head *head_base)
+{
+	struct wayland_backend *b = to_wayland_backend(output_base->compositor);
+	struct wayland_output *output = to_wayland_output(output_base);
+	struct wayland_head *head = to_wayland_head(head_base);
+
+	if (!wl_list_empty(&output->base.head_list))
+		return -1;
+
+	if (head->parent_output) {
+		if (wayland_output_setup_for_parent_output(output,
+							   head->parent_output) < 0)
+			return -1;
+	} else if (b->fullscreen) {
+		if (wayland_output_setup_fullscreen(output, head) < 0)
+			return -1;
+	} else {
+		/* A floating window, nothing to do. */
+	}
+
+	return 0;
+}
+
+static void
+wayland_output_detach_head(struct weston_output *output_base,
+			   struct weston_head *head)
+{
+	struct wayland_output *output = to_wayland_output(output_base);
+
+	/* Rely on the disable hook if the output was enabled. We do not
+	 * support cloned heads, so detaching is guaranteed to disable the
+	 * output.
+	 */
+	if (output->base.enabled)
+		return;
+
+	/* undo setup fullscreen */
+	if (output->parent.surface)
+		wayland_backend_destroy_output_surface(output);
+}
+
+static struct weston_output *
+wayland_output_create(struct weston_compositor *compositor, const char *name)
 {
 	struct wayland_output *output;
+	char *title;
+
+	/* name can't be NULL. */
+	assert(name);
 
 	output = zalloc(sizeof *output);
 	if (output == NULL) {
@@ -1247,33 +1355,98 @@ wayland_output_create_common(void)
 		return NULL;
 	}
 
+	if (asprintf(&title, "%s - %s", WINDOW_TITLE, name) < 0) {
+		free(output);
+		return NULL;
+	}
+	output->title = title;
+
+	weston_output_init(&output->base, compositor, name);
+
 	output->base.destroy = wayland_output_destroy;
 	output->base.disable = wayland_output_disable;
 	output->base.enable = wayland_output_enable;
+	output->base.attach_head = wayland_output_attach_head;
+	output->base.detach_head = wayland_output_detach_head;
 
-	return output;
+	weston_compositor_add_pending_output(&output->base, compositor);
+
+	return &output->base;
+}
+
+static struct wayland_head *
+wayland_head_create(struct weston_compositor *compositor, const char *name)
+{
+	struct wayland_head *head;
+
+	assert(name);
+
+	head = zalloc(sizeof *head);
+	if (!head)
+		return NULL;
+
+	weston_head_init(&head->base, name);
+	weston_head_set_connection_status(&head->base, true);
+	weston_compositor_add_head(compositor, &head->base);
+
+	return head;
 }
 
 static int
-wayland_output_create(struct weston_compositor *compositor, const char *name)
+wayland_head_create_windowed(struct weston_compositor *compositor,
+			     const char *name)
 {
-	struct wayland_output *output = wayland_output_create_common();
-
-	/* name can't be NULL. */
-	assert(name);
-
-	output->base.name = strdup(name);
-
-	weston_output_init(&output->base, compositor);
-	weston_compositor_add_pending_output(&output->base, compositor);
+	 if (!wayland_head_create(compositor, name))
+		return -1;
 
 	return 0;
 }
 
+static int
+wayland_head_create_for_parent_output(struct weston_compositor *compositor,
+				      struct wayland_parent_output *poutput)
+{
+	struct wayland_head *head;
+	char name[100];
+	int ret;
+
+	ret = snprintf(name, sizeof(name), "wlparent-%d", poutput->id);
+	if (ret < 1 || (unsigned)ret >= sizeof(name))
+		return -1;
+
+	head = wayland_head_create(compositor, name);
+	if (!head)
+		return -1;
+
+	assert(!poutput->head);
+	head->parent_output = poutput;
+	poutput->head = head;
+
+	weston_head_set_monitor_strings(&head->base,
+					poutput->physical.make,
+					poutput->physical.model, NULL);
+	weston_head_set_physical_size(&head->base,
+				      poutput->physical.width,
+				      poutput->physical.height);
+
+	return 0;
+}
+
+static void
+wayland_head_destroy(struct wayland_head *head)
+{
+	if (head->parent_output)
+		head->parent_output->head = NULL;
+
+	weston_head_release(&head->base);
+	free(head);
+}
+
 static int
 wayland_output_set_size(struct weston_output *base, int width, int height)
 {
 	struct wayland_output *output = to_wayland_output(base);
+	struct weston_head *head;
 	int output_width, output_height;
 
 	/* We can only be called once. */
@@ -1294,6 +1467,13 @@ wayland_output_set_size(struct weston_output *base, int width, int height)
 		return -1;
 	}
 
+	wl_list_for_each(head, &output->base.head_list, output_link) {
+		weston_head_set_monitor_strings(head, "wayland", "none", NULL);
+
+		/* XXX: Calculate proper size. */
+		weston_head_set_physical_size(head, width, height);
+	}
+
 	output_width = width * output->base.scale;
 	output_height = height * output->base.scale;
 
@@ -1303,30 +1483,19 @@ wayland_output_set_size(struct weston_output *base, int width, int height)
 	output->mode.width = output_width;
 	output->mode.height = output_height;
 	output->mode.refresh = 60000;
-	output->scale = output->base.scale;
-	wl_list_init(&output->base.mode_list);
 	wl_list_insert(&output->base.mode_list, &output->mode.link);
 
 	output->base.current_mode = &output->mode;
-	output->base.make = "wayland";
-	output->base.model = "none";
-
-	/* XXX: Calculate proper size. */
-	output->base.mm_width = width;
-	output->base.mm_height = height;
 
 	return 0;
 }
 
 static int
-wayland_output_create_for_parent_output(struct wayland_backend *b,
-					struct wayland_parent_output *poutput)
+wayland_output_setup_for_parent_output(struct wayland_output *output,
+				       struct wayland_parent_output *poutput)
 {
-	struct wayland_output *output = wayland_output_create_common();
 	struct weston_mode *mode;
 
-	output->base.name = strdup("wlparent");
-
 	if (poutput->current_mode) {
 		mode = poutput->current_mode;
 	} else if (poutput->preferred_mode) {
@@ -1336,53 +1505,38 @@ wayland_output_create_for_parent_output(struct wayland_backend *b,
 				    struct weston_mode, link);
 	} else {
 		weston_log("No valid modes found. Skipping output.\n");
-		goto out;
+		return -1;
 	}
 
-	weston_output_init(&output->base, b->compositor);
-
 	output->base.scale = 1;
 	output->base.transform = WL_OUTPUT_TRANSFORM_NORMAL;
 
-	if (wayland_output_set_size(&output->base, mode->width, mode->height) < 0)
-		goto out;
-
-	output->mode = *mode;
 	output->parent.output = poutput->global;
 
-	output->base.make = poutput->physical.make;
-	output->base.model = poutput->physical.model;
-
-	wl_list_init(&output->base.mode_list);
 	wl_list_insert_list(&output->base.mode_list, &poutput->mode_list);
 	wl_list_init(&poutput->mode_list);
 
-	weston_compositor_add_pending_output(&output->base, b->compositor);
+	/* No other mode should have CURRENT already. */
+	mode->flags |= WL_OUTPUT_MODE_CURRENT;
+	output->base.current_mode = mode;
+
+	/* output->mode is unused in this path. */
 
 	return 0;
-
-out:
-	free(output->name);
-	free(output);
-
-	return -1;
 }
 
 static int
-wayland_output_create_fullscreen(struct wayland_backend *b)
+wayland_output_setup_fullscreen(struct wayland_output *output,
+				struct wayland_head *head)
 {
-	struct wayland_output *output = wayland_output_create_common();
+	struct wayland_backend *b = to_wayland_backend(output->base.compositor);
 	int width = 0, height = 0;
 
-	output->base.name = strdup("wayland-fullscreen");
-
-	weston_output_init(&output->base, b->compositor);
-
 	output->base.scale = 1;
 	output->base.transform = WL_OUTPUT_TRANSFORM_NORMAL;
 
 	if (wayland_backend_create_output_surface(output) < 0)
-		goto err_surface;
+		return -1;
 
 	/* What should size be set if conditional is false? */
 	if (b->parent.xdg_shell || b->parent.shell) {
@@ -1402,15 +1556,15 @@ wayland_output_create_fullscreen(struct wayland_backend *b)
 	if (wayland_output_set_size(&output->base, width, height) < 0)
 		goto err_set_size;
 
-	weston_compositor_add_pending_output(&output->base, b->compositor);
+	/* The head is not attached yet, so set_size did not set these. */
+	weston_head_set_monitor_strings(&head->base, "wayland", "none", NULL);
+	/* XXX: Calculate proper size. */
+	weston_head_set_physical_size(&head->base, width, height);
 
 	return 0;
 
 err_set_size:
 	wayland_backend_destroy_output_surface(output);
-err_surface:
-	free(output->name);
-	free(output);
 
 	return -1;
 }
@@ -1483,6 +1637,13 @@ input_handle_pointer_enter(void *data, struct wl_pointer *pointer,
 	enum theme_location location;
 	double x, y;
 
+	if (!surface) {
+		input->output = NULL;
+		input->has_focus = false;
+		notify_pointer_focus(&input->base, NULL, 0, 0);
+		return;
+	}
+
 	x = wl_fixed_to_double(fixed_x);
 	y = wl_fixed_to_double(fixed_y);
 
@@ -1548,6 +1709,7 @@ input_handle_motion(void *data, struct wl_pointer *pointer,
 	enum theme_location location;
 	bool want_frame = false;
 	double x, y;
+	struct timespec ts;
 
 	if (!input->output)
 		return;
@@ -1585,7 +1747,8 @@ input_handle_motion(void *data, struct wl_pointer *pointer,
 	}
 
 	if (location == THEME_LOCATION_CLIENT_AREA) {
-		notify_motion_absolute(&input->base, time, x, y);
+		timespec_from_msec(&ts, time);
+		notify_motion_absolute(&input->base, &ts, x, y);
 		want_frame = true;
 	}
 
@@ -1600,6 +1763,7 @@ input_handle_button(void *data, struct wl_pointer *pointer,
 {
 	struct wayland_input *input = data;
 	enum theme_location location;
+	struct timespec ts;
 
 	if (!input->output)
 		return;
@@ -1638,7 +1802,8 @@ input_handle_button(void *data, struct wl_pointer *pointer,
 	}
 
 	if (location == THEME_LOCATION_CLIENT_AREA) {
-		notify_button(&input->base, time, button, state);
+		timespec_from_msec(&ts, time);
+		notify_button(&input->base, &ts, button, state);
 		if (input->seat_version < WL_POINTER_FRAME_SINCE_VERSION)
 			notify_pointer_frame(&input->base);
 	}
@@ -1650,9 +1815,11 @@ input_handle_axis(void *data, struct wl_pointer *pointer,
 {
 	struct wayland_input *input = data;
 	struct weston_pointer_axis_event weston_event;
+	struct timespec ts;
 
 	weston_event.axis = axis;
 	weston_event.value = wl_fixed_to_double(value);
+	weston_event.has_discrete = false;
 
 	if (axis == WL_POINTER_AXIS_VERTICAL_SCROLL &&
 	    input->vert.has_discrete) {
@@ -1666,7 +1833,9 @@ input_handle_axis(void *data, struct wl_pointer *pointer,
 		input->horiz.has_discrete = false;
 	}
 
-	notify_axis(&input->base, time, &weston_event);
+	timespec_from_msec(&ts, time);
+
+	notify_axis(&input->base, &ts, &weston_event);
 
 	if (input->seat_version < WL_POINTER_FRAME_SINCE_VERSION)
 		notify_pointer_frame(&input->base);
@@ -1695,11 +1864,14 @@ input_handle_axis_stop(void *data, struct wl_pointer *pointer,
 {
 	struct wayland_input *input = data;
 	struct weston_pointer_axis_event weston_event;
+	struct timespec ts;
 
 	weston_event.axis = axis;
 	weston_event.value = 0;
 
-	notify_axis(&input->base, time, &weston_event);
+	timespec_from_msec(&ts, time);
+
+	notify_axis(&input->base, &ts, &weston_event);
 }
 
 static void
@@ -1806,6 +1978,11 @@ input_handle_keyboard_enter(void *data,
 			weston_output_schedule_repaint(&focus->base);
 	}
 
+	if (!surface) {
+		input->keyboard_focus = NULL;
+		return;
+	}
+
 	input->keyboard_focus = wl_surface_get_user_data(surface);
 	input->keyboard_focus->keyboard_count++;
 
@@ -1853,9 +2030,15 @@ input_handle_key(void *data, struct wl_keyboard *keyboard,
 		 uint32_t serial, uint32_t time, uint32_t key, uint32_t state)
 {
 	struct wayland_input *input = data;
+	struct timespec ts;
+
+	if (!input->keyboard_focus)
+		return;
+
+	timespec_from_msec(&ts, time);
 
 	input->key_serial = serial;
-	notify_key(&input->base, time, key,
+	notify_key(&input->base, &ts, key,
 		   state ? WL_KEYBOARD_KEY_STATE_PRESSED :
 			   WL_KEYBOARD_KEY_STATE_RELEASED,
 		   input->keyboard_state_update);
@@ -1919,10 +2102,13 @@ input_handle_touch_down(void *data, struct wl_touch *wl_touch,
 	bool first_touch;
 	int32_t fx, fy;
 	double x, y;
+	struct timespec ts;
 
 	x = wl_fixed_to_double(fixed_x);
 	y = wl_fixed_to_double(fixed_y);
 
+	timespec_from_msec(&ts, time);
+
 	first_touch = (input->touch_points == 0);
 	input->touch_points++;
 
@@ -1960,7 +2146,7 @@ input_handle_touch_down(void *data, struct wl_touch *wl_touch,
 
 	weston_output_transform_coordinate(&output->base, x, y, &x, &y);
 
-	notify_touch(&input->base, time, id, x, y, WL_TOUCH_DOWN);
+	notify_touch(input->touch_device, &ts, id, x, y, WL_TOUCH_DOWN);
 	input->touch_active = true;
 }
 
@@ -1971,6 +2157,9 @@ input_handle_touch_up(void *data, struct wl_touch *wl_touch,
 	struct wayland_input *input = data;
 	struct wayland_output *output = input->touch_focus;
 	bool active = input->touch_active;
+	struct timespec ts;
+
+	timespec_from_msec(&ts, time);
 
 	input->touch_points--;
 	if (input->touch_points == 0) {
@@ -1998,7 +2187,7 @@ input_handle_touch_up(void *data, struct wl_touch *wl_touch,
 	}
 
 	if (active)
-		notify_touch(&input->base, time, id, 0, 0, WL_TOUCH_UP);
+		notify_touch(input->touch_device, &ts, id, 0, 0, WL_TOUCH_UP);
 }
 
 static void
@@ -2010,9 +2199,11 @@ input_handle_touch_motion(void *data, struct wl_touch *wl_touch,
 	struct wayland_output *output = input->touch_focus;
 	int32_t fx, fy;
 	double x, y;
+	struct timespec ts;
 
 	x = wl_fixed_to_double(fixed_x);
 	y = wl_fixed_to_double(fixed_y);
+	timespec_from_msec(&ts, time);
 
 	if (!output || !input->touch_active)
 		return;
@@ -2025,7 +2216,7 @@ input_handle_touch_motion(void *data, struct wl_touch *wl_touch,
 
 	weston_output_transform_coordinate(&output->base, x, y, &x, &y);
 
-	notify_touch(&input->base, time, id, x, y, WL_TOUCH_MOTION);
+	notify_touch(input->touch_device, &ts, id, x, y, WL_TOUCH_MOTION);
 }
 
 static void
@@ -2036,7 +2227,7 @@ input_handle_touch_frame(void *data, struct wl_touch *wl_touch)
 	if (!input->touch_focus || !input->touch_active)
 		return;
 
-	notify_touch_frame(&input->base);
+	notify_touch_frame(input->touch_device);
 }
 
 static void
@@ -2047,7 +2238,7 @@ input_handle_touch_cancel(void *data, struct wl_touch *wl_touch)
 	if (!input->touch_focus || !input->touch_active)
 		return;
 
-	notify_touch_cancel(&input->base);
+	notify_touch_cancel(input->touch_device);
 }
 
 static const struct wl_touch_listener touch_listener = {
@@ -2059,6 +2250,22 @@ static const struct wl_touch_listener touch_listener = {
 };
 
 
+static struct weston_touch_device *
+create_touch_device(struct wayland_input *input)
+{
+	struct weston_touch_device *touch_device;
+	char str[128];
+
+	/* manufacture a unique'ish name */
+	snprintf(str, sizeof str, "wayland-touch[%u]",
+		 wl_proxy_get_id((struct wl_proxy *)input->parent.seat));
+
+	touch_device = weston_touch_create_touch_device(input->base.touch_state,
+							str, NULL, NULL);
+
+	return touch_device;
+}
+
 static void
 input_handle_capabilities(void *data, struct wl_seat *seat,
 		          enum wl_seat_capability caps)
@@ -2100,7 +2307,10 @@ input_handle_capabilities(void *data, struct wl_seat *seat,
 		wl_touch_add_listener(input->parent.touch,
 				      &touch_listener, input);
 		weston_seat_init_touch(&input->base);
+		input->touch_device = create_touch_device(input);
 	} else if (!(caps & WL_SEAT_CAPABILITY_TOUCH) && input->parent.touch) {
+		weston_touch_device_destroy(input->touch_device);
+		input->touch_device = NULL;
 		if (input->seat_version >= WL_TOUCH_RELEASE_SINCE_VERSION)
 			wl_touch_release(input->parent.touch);
 		else
@@ -2194,16 +2404,32 @@ find_mode(struct wl_list *list, int32_t width, int32_t height, uint32_t refresh)
 	return mode;
 }
 
+static struct weston_output *
+wayland_parent_output_get_enabled_output(struct wayland_parent_output *poutput)
+{
+	struct wayland_head *head = poutput->head;
+
+	if (!head)
+		return NULL;
+
+	if (!weston_head_is_enabled(&head->base))
+		return NULL;
+
+	return weston_head_get_output(&head->base);
+}
+
 static void
 wayland_parent_output_mode(void *data, struct wl_output *wl_output_proxy,
 			   uint32_t flags, int32_t width, int32_t height,
 			   int32_t refresh)
 {
 	struct wayland_parent_output *output = data;
+	struct weston_output *enabled_output;
 	struct weston_mode *mode;
 
-	if (output->output) {
-		mode = find_mode(&output->output->base.mode_list,
+	enabled_output = wayland_parent_output_get_enabled_output(output);
+	if (enabled_output) {
+		mode = find_mode(&enabled_output->mode_list,
 				 width, height, refresh);
 		if (!mode)
 			return;
@@ -2226,6 +2452,24 @@ static const struct wl_output_listener output_listener = {
 	wayland_parent_output_mode
 };
 
+static void
+output_sync_callback(void *data, struct wl_callback *callback, uint32_t unused)
+{
+	struct wayland_parent_output *output = data;
+
+	assert(output->sync_cb == callback);
+	wl_callback_destroy(callback);
+	output->sync_cb = NULL;
+
+	assert(output->backend->sprawl_across_outputs);
+
+	wayland_head_create_for_parent_output(output->backend->compositor, output);
+}
+
+static const struct wl_callback_listener output_sync_listener = {
+	output_sync_callback
+};
+
 static void
 wayland_backend_register_output(struct wayland_backend *b, uint32_t id)
 {
@@ -2235,6 +2479,7 @@ wayland_backend_register_output(struct wayland_backend *b, uint32_t id)
 	if (!output)
 		return;
 
+	output->backend = b;
 	output->id = id;
 	output->global = wl_registry_bind(b->parent.registry, id,
 					  &wl_output_interface, 1);
@@ -2252,8 +2497,9 @@ wayland_backend_register_output(struct wayland_backend *b, uint32_t id)
 	wl_list_insert(&b->parent.output_list, &output->link);
 
 	if (b->sprawl_across_outputs) {
-		wl_display_roundtrip(b->parent.wl_display);
-		wayland_output_create_for_parent_output(b, output);
+		output->sync_cb = wl_display_sync(b->parent.wl_display);
+		wl_callback_add_listener(output->sync_cb,
+					 &output_sync_listener, output);
 	}
 }
 
@@ -2262,8 +2508,11 @@ wayland_parent_output_destroy(struct wayland_parent_output *output)
 {
 	struct weston_mode *mode, *next;
 
-	if (output->output)
-		wayland_output_destroy(&output->output->base);
+	if (output->sync_cb)
+		wl_callback_destroy(output->sync_cb);
+
+	if (output->head)
+		wayland_head_destroy(output->head);
 
 	wl_output_destroy(output->global);
 	free(output->physical.make);
@@ -2273,6 +2522,9 @@ wayland_parent_output_destroy(struct wayland_parent_output *output)
 		wl_list_remove(&mode->link);
 		free(mode);
 	}
+
+	wl_list_remove(&output->link);
+	free(output);
 }
 
 static void
@@ -2294,7 +2546,8 @@ registry_handle_global(void *data, struct wl_registry *registry, uint32_t name,
 	if (strcmp(interface, "wl_compositor") == 0) {
 		b->parent.compositor =
 			wl_registry_bind(registry, name,
-					 &wl_compositor_interface, 1);
+					 &wl_compositor_interface,
+					 MIN(version, 4));
 	} else if (strcmp(interface, "zxdg_shell_v6") == 0) {
 		b->parent.xdg_shell =
 			wl_registry_bind(registry, name,
@@ -2324,9 +2577,9 @@ registry_handle_global_remove(void *data, struct wl_registry *registry,
 			      uint32_t name)
 {
 	struct wayland_backend *b = data;
-	struct wayland_parent_output *output;
+	struct wayland_parent_output *output, *next;
 
-	wl_list_for_each(output, &b->parent.output_list, link)
+	wl_list_for_each_safe(output, next, &b->parent.output_list, link)
 		if (output->id == name)
 			wayland_parent_output_destroy(output);
 }
@@ -2360,20 +2613,19 @@ wayland_backend_handle_event(int fd, uint32_t mask, void *data)
 	return count;
 }
 
-static void
-wayland_restore(struct weston_compositor *ec)
-{
-}
-
 static void
 wayland_destroy(struct weston_compositor *ec)
 {
 	struct wayland_backend *b = to_wayland_backend(ec);
+	struct weston_head *base, *next;
 
 	wl_event_source_remove(b->parent.wl_source);
 
 	weston_compositor_shutdown(ec);
 
+	wl_list_for_each_safe(base, next, &ec->head_list, compositor_link)
+		wayland_head_destroy(to_wayland_head(base));
+
 	if (b->parent.shm)
 		wl_shm_destroy(b->parent.shm);
 
@@ -2436,8 +2688,8 @@ create_cursor(struct wayland_backend *b,
 }
 
 static void
-fullscreen_binding(struct weston_keyboard *keyboard, uint32_t time,
-		   uint32_t key, void *data)
+fullscreen_binding(struct weston_keyboard *keyboard,
+		   const struct timespec *time, uint32_t key, void *data)
 {
 	struct wayland_backend *b = data;
 	struct wayland_input *input = NULL;
@@ -2470,12 +2722,17 @@ wayland_backend_create(struct weston_compositor *compositor,
 		return NULL;
 
 	b->compositor = compositor;
+	compositor->backend = &b->base;
+
 	if (weston_compositor_set_presentation_clock_software(compositor) < 0)
 		goto err_compositor;
 
 	b->parent.wl_display = wl_display_connect(new_config->display_name);
 	if (b->parent.wl_display == NULL) {
-		weston_log("failed to create display: %m\n");
+		weston_log("Error: Failed to connect to parent Wayland compositor: %m\n");
+		weston_log_continue(STAMP_SPACE "display option: %s, WAYLAND_DISPLAY=%s\n",
+				    new_config->display_name ?: "(none)",
+				    getenv("WAYLAND_DISPLAY") ?: "(not set)");
 		goto err_compositor;
 	}
 
@@ -2523,7 +2780,7 @@ wayland_backend_create(struct weston_compositor *compositor,
 	}
 
 	b->base.destroy = wayland_destroy;
-	b->base.restore = wayland_restore;
+	b->base.create_output = wayland_output_create;
 
 	loop = wl_display_get_event_loop(compositor->wl_display);
 
@@ -2542,7 +2799,6 @@ wayland_backend_create(struct weston_compositor *compositor,
 			           "support failed.\n");
 	}
 
-	compositor->backend = &b->base;
 	return b;
 err_display:
 	wl_display_disconnect(b->parent.wl_display);
@@ -2569,7 +2825,7 @@ wayland_backend_destroy(struct wayland_backend *b)
 
 static const struct weston_windowed_output_api windowed_api = {
 	wayland_output_set_size,
-	wayland_output_create,
+	wayland_head_create_windowed,
 };
 
 static void
@@ -2606,14 +2862,14 @@ weston_backend_init(struct weston_compositor *compositor,
 		wl_display_roundtrip(b->parent.wl_display);
 
 		wl_list_for_each(poutput, &b->parent.output_list, link)
-			wayland_output_create_for_parent_output(b, poutput);
+			wayland_head_create_for_parent_output(compositor, poutput);
 
 		return 0;
 	}
 
 	if (new_config.fullscreen) {
-		if (wayland_output_create_fullscreen(b) < 0) {
-			weston_log("Unable to create a fullscreen output.\n");
+		if (!wayland_head_create(compositor, "wayland-fullscreen")) {
+			weston_log("Unable to create a fullscreen head.\n");
 			goto err_outputs;
 		}
 
