@@ -35,7 +35,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
-#include <sys/sysmacros.h>
 #include <systemd/sd-login.h>
 #include <unistd.h>
 
@@ -45,6 +44,14 @@
 
 #define DRM_MAJOR 226
 
+/* major()/minor() */
+#ifdef MAJOR_IN_MKDEV
+#include <sys/mkdev.h>
+#endif
+#ifdef MAJOR_IN_SYSMACROS
+#include <sys/sysmacros.h>
+#endif
+
 struct launcher_logind {
 	struct weston_launcher base;
 	struct weston_compositor *compositor;
@@ -216,6 +223,7 @@ launcher_logind_close(struct weston_launcher *launcher, int fd)
 	int r;
 
 	r = fstat(fd, &st);
+	close(fd);
 	if (r < 0) {
 		weston_log("logind: cannot fstat fd: %m\n");
 		return;
@@ -230,11 +238,6 @@ launcher_logind_close(struct weston_launcher *launcher, int fd)
 				     minor(st.st_rdev));
 }
 
-static void
-launcher_logind_restore(struct weston_launcher *launcher)
-{
-}
-
 static int
 launcher_logind_activate_vt(struct weston_launcher *launcher, int vt)
 {
@@ -374,7 +377,6 @@ static void
 disconnected_dbus(struct launcher_logind *wl)
 {
 	weston_log("logind: dbus connection lost, exiting..\n");
-	launcher_logind_restore(&wl->base);
 	exit(-1);
 }
 
@@ -395,7 +397,6 @@ session_removed(struct launcher_logind *wl, DBusMessage *m)
 
 	if (!strcmp(name, wl->sid)) {
 		weston_log("logind: our session got closed, exiting..\n");
-		launcher_logind_restore(&wl->base);
 		exit(-1);
 	}
 }
@@ -761,17 +762,20 @@ launcher_logind_connect(struct weston_launcher **out, struct weston_compositor *
 		free(t);
 		goto err_session;
 	}
-	free(t);
 
-	r = weston_sd_session_get_vt(wl->sid, &wl->vtnr);
-	if (r < 0) {
-		weston_log("logind: session not running on a VT\n");
-		goto err_session;
-	} else if (tty > 0 && wl->vtnr != (unsigned int )tty) {
-		weston_log("logind: requested VT --tty=%d differs from real session VT %u\n",
-			   tty, wl->vtnr);
-		r = -EINVAL;
-		goto err_session;
+	r = strcmp(t, "seat0");
+	free(t);
+	if (r == 0) {
+		r = weston_sd_session_get_vt(wl->sid, &wl->vtnr);
+		if (r < 0) {
+			weston_log("logind: session not running on a VT\n");
+			goto err_session;
+		} else if (tty > 0 && wl->vtnr != (unsigned int )tty) {
+			weston_log("logind: requested VT --tty=%d differs from real session VT %u\n",
+				   tty, wl->vtnr);
+			r = -EINVAL;
+			goto err_session;
+		}
 	}
 
 	loop = wl_display_get_event_loop(compositor->wl_display);
@@ -844,6 +848,5 @@ const struct launcher_interface launcher_logind_iface = {
 	.open = launcher_logind_open,
 	.close = launcher_logind_close,
 	.activate_vt = launcher_logind_activate_vt,
-	.restore = launcher_logind_restore,
 	.get_vt = launcher_logind_get_vt,
 };
